from cpl_core.database.context import DatabaseContextABC
from cpl_discord.service import DiscordBotServiceABC

from bot_data.abc.achievement_repository_abc import AchievementRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.model.achievement import Achievement
from bot_data.model.user_role_enum import UserRoleEnum
from bot_graphql.abc.query_abc import QueryABC
from modules.permission.service.permission_service import PermissionService


class AchievementMutation(QueryABC):
    def __init__(
        self,
        servers: ServerRepositoryABC,
        achievements: AchievementRepositoryABC,
        bot: DiscordBotServiceABC,
        db: DatabaseContextABC,
        permissions: PermissionService,
    ):
        QueryABC.__init__(self, "AchievementMutation")

        self._servers = servers
        self._achievements = achievements
        self._bot = bot
        self._db = db
        self._permissions = permissions

        self.set_field("createAchievement", self.resolve_create_achievement)
        self.set_field("updateAchievement", self.resolve_update_achievement)
        self.set_field("deleteAchievement", self.resolve_delete_achievement)

    def resolve_create_achievement(self, *_, input: dict):
        server = self._servers.get_server_by_id(input["serverId"])
        self._can_user_mutate_data(server, UserRoleEnum.admin)

        achievement = Achievement(
            input["name"],
            input["description"],
            input["attribute"],
            input["operator"],
            input["value"],
            server,
        )
        self._achievements.add_achievement(achievement)
        self._db.save_changes()

        def get_new_achievement(a: Achievement):
            return (
                a.name == achievement.name
                and a.description == achievement.description
                and a.attribute == achievement.attribute
                and a.operator == achievement.operator
                and a.value == achievement.value
                and a.server.id == server.id
            )

        return self._achievements.get_achievements_by_server_id(achievement.server.id).where(get_new_achievement).last()

    def resolve_update_achievement(self, *_, input: dict):
        achievement = self._achievements.get_achievement_by_id(input["id"])
        self._can_user_mutate_data(achievement.server, UserRoleEnum.moderator)

        achievement.name = input["name"] if "name" in input else achievement.name
        achievement.description = input["description"] if "description" in input else achievement.description
        achievement.attribute = input["attribute"] if "attribute" in input else achievement.attribute
        achievement.operator = input["operator"] if "operator" in input else achievement.operator
        achievement.value = input["value"] if "value" in input else achievement.value

        self._achievements.update_achievement(achievement)
        self._db.save_changes()

        achievement = self._achievements.get_achievement_by_id(input["id"])
        return achievement

    def resolve_delete_achievement(self, *_, id: int):
        achievement = self._achievements.get_achievement_by_id(id)
        self._can_user_mutate_data(achievement.server, UserRoleEnum.admin)

        joins = self._achievements.get_user_got_achievements_by_achievement_id(id)
        for join in joins:
            self._achievements.delete_user_got_achievement(join)

        self._achievements.delete_achievement(achievement)
        self._db.save_changes()

        return achievement
