import { Injectable } from "@angular/core";
import { MenuItem } from "primeng/api";
import { BehaviorSubject, forkJoin, Observable } from "rxjs";
import { AuthRoles } from "../../models/auth/auth-roles.enum";
import { AuthService } from "../auth/auth.service";
import { TranslateService } from "@ngx-translate/core";
import { NavigationEnd, Router } from "@angular/router";
import { ThemeService } from "../theme/theme.service";
import { Server } from "../../models/data/server.model";
import { UserDTO } from "../../models/auth/auth-user.dto";
import { ServerService } from "../server.service";
import { HasServerFeatureFlagQuery, PossibleFeatureFlagsQuery, Query } from "../../models/graphql/query.model";
import { Queries } from "../../models/graphql/queries.model";
import { DataService } from "../data/data.service";
import { FeatureFlag } from "../../models/config/feature-flags.model";

@Injectable({
  providedIn: "root"
})
export class SidebarService {

  isSidebarOpen: boolean = true;
  menuItems$ = new BehaviorSubject<MenuItem[]>(new Array<MenuItem>());
  server!: Server | undefined;

  dashboard: MenuItem = {};
  serverDashboard: MenuItem = {};
  serverProfile: MenuItem = {};
  serverMembers: MenuItem = {};
  serverAutoRoles: MenuItem = {};
  serverLevels: MenuItem = {};
  serverAchievements: MenuItem = {};
  serverShortRoleNames: MenuItem = {};
  serverConfig: MenuItem = {};
  serverMenu: MenuItem = {};
  adminConfig: MenuItem = {};
  adminUsers: MenuItem = {};
  adminMenu: MenuItem = {};

  featureFlags: FeatureFlag[] = [];

  constructor(
    private themeService: ThemeService,
    private authService: AuthService,
    private translateService: TranslateService,
    private router: Router,
    private serverService: ServerService,
    private data: DataService
  ) {
    this.themeService.isSidebarOpen$.subscribe(value => {
      this.isSidebarOpen = value;
      this.setMenu(true);
    });

    this.translateService.onLangChange.subscribe(_ => {
      this.setMenu(true);
    });

    this.serverService.server$.subscribe(server => {
      this.server = server;
      if (server) {
        this.setMenu(true);
      } else {
        this.setMenu(false);
      }
    });
  }

  async buildMenu(user: UserDTO | null, hasPermission: boolean, isTechnician: boolean = false) {
    this.dashboard = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.dashboard") : "",
      icon: "pi pi-th-large",
      routerLink: "dashboard"
    };
    this.serverDashboard = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.dashboard") : "",
      icon: "pi pi-th-large",
      routerLink: `server/${this.server?.id}`
    };
    this.serverProfile = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.profile") : "",
      icon: "pi pi-id-card",
      routerLink: `server/${this.server?.id}/members/${user?.id}`
    };
    this.serverMembers = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.members") : "",
      icon: "pi pi-users",
      visible: true,
      routerLink: `server/${this.server?.id}/members`
    };

    this.serverAutoRoles = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.auto_roles") : "",
      icon: "pi pi-sitemap",
      visible: true,
      routerLink: `server/${this.server?.id}/auto-roles`
    };

    this.serverLevels = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.levels") : "",
      icon: "pi pi-book",
      visible: true,
      routerLink: `server/${this.server?.id}/levels`
    };

    this.serverAchievements = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.achievements") : "",
      icon: "pi pi-angle-double-up",
      visible: true,
      routerLink: `server/${this.server?.id}/achievements`
    };

    this.serverShortRoleNames = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.short_role_names") : "",
      icon: "pi pi-list",
      visible: true,
      routerLink: `server/${this.server?.id}/short-role-names`
    };

    this.serverConfig = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.server.configuration") : "",
      icon: "pi pi-cog",
      visible: true,
      routerLink: `server/${this.server?.id}/config`
    };

    this.serverMenu = {
      label: this.isSidebarOpen ? this.server?.name : "",
      icon: "pi pi-server",
      visible: false,
      expanded: true,
      items: [this.serverDashboard, this.serverProfile, this.serverMembers, this.serverAutoRoles, this.serverLevels, this.serverAchievements, this.serverShortRoleNames, this.serverConfig]
    };
    this.adminConfig = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.config") : "",
      visible: hasPermission || isTechnician,
      icon: "pi pi-cog",
      routerLink: "/admin/settings"
    };
    this.adminUsers = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.auth_user_list") : "",
      visible: hasPermission,
      icon: "pi pi-user-edit",
      routerLink: "/admin/users"
    };
    this.adminMenu = {
      label: this.isSidebarOpen ? this.translateService.instant("sidebar.administration") : "",
      icon: "pi pi-cog",
      visible: hasPermission || isTechnician,
      expanded: true,
      items: [this.adminConfig, this.adminUsers]
    };
  }

  setMenu(build: boolean = false) {
    const server = this.server;

    if (server) {
      this.featureFlags = [];
      this.data.query<PossibleFeatureFlagsQuery>("{possibleFeatureFlags}"
      ).subscribe(data => {
        let observables: Observable<HasServerFeatureFlagQuery>[] = [];
        data.possibleFeatureFlags.forEach(flag => {
          observables.push(
            this.data.query<HasServerFeatureFlagQuery>(Queries.hasServerFeatureFlag, {
                filter: { id: server.id },
                flag: flag
              },
              function(data: Query) {
                return data.servers[0];
              }
            )
          );
        });

        forkJoin(observables).subscribe(data => {
          data.forEach(flag => {
            if (!flag.hasFeatureFlag.value) {
              return;
            }
            this.featureFlags.push(flag.hasFeatureFlag);
          });
          this._setMenu(build);
        });
      });
    } else {
      this._setMenu(build);
    }
  }

  private _setMenu(build: boolean = false) {
    this.authService.hasUserPermission(AuthRoles.Admin).then(async hasPermission => {
      let authUser = await this.authService.getLoggedInUser();
      let user: UserDTO | null = authUser?.users?.find(u => u.server == this.server?.id) ?? null;
      let isTechnician = (authUser?.users?.map(u => u.isTechnician).filter(u => u) ?? []).length > 0;
      let isTechnicianAndFullAccessActive = this.hasFeature("TechnicianFullAccess") && isTechnician;

      if (build || this.menuItems$.value.length == 0) {
        await this.buildMenu(user, hasPermission, isTechnician);
      }

      if (this.server) {
        this.serverMenu.visible = true;
        this.serverMembers.visible = isTechnicianAndFullAccessActive || user?.isModerator;
        this.serverAutoRoles.visible = isTechnicianAndFullAccessActive || this.hasFeature("AutoRoleModule") && user?.isModerator;
        this.serverLevels.visible = isTechnicianAndFullAccessActive || this.hasFeature("LevelModule") && user?.isModerator;
        this.serverAchievements.visible = isTechnicianAndFullAccessActive || this.hasFeature("AchievementsModule") && user?.isModerator;
        this.serverShortRoleNames.visible = isTechnicianAndFullAccessActive || this.hasFeature("ShortRoleName") && user?.isAdmin;

        this.serverConfig.visible = isTechnicianAndFullAccessActive || user?.isAdmin;
      } else {
        this.serverMenu.visible = false;
      }

      let menuItems: MenuItem[] = [
        this.dashboard,
        this.serverMenu,
        this.adminMenu
      ];
      this.menuItems$.next(menuItems);
    });
  }

  public hasFeature(key: string): boolean {
    const flag = this.featureFlags.filter(flag => flag.key == key);
    if (flag.length == 0) {
      return false;
    }
    return flag[0].value;
  }
}
